
/*
 * blockcipher.c --
 *
 *	Implementation of procedures realizing the generic parts
 *	of block ciphers.
 *
 * Copyright (c) 1995 Andreas Kupries (aku@kisters.de)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: blockcipher.c,v 1.1 1996/02/11 00:03:12 aku Exp $
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include <blobXInt.h>

/*
 * declaration of internal types
 */

/*
 * Every command realizing a block cipher is given
 * an instance of the following structure as clientdata.
 * It enables the generic procedures to call the realization
 * of the special cipher.
 */

typedef struct BcData {
  int                 keySize;     /* (minimal) size of key (in bytes) */
  int                 blockSize;   /* size of single block (in bytes)  */
  BlobX_BCScheduleKey schedule;    /* proc to compile external key     */
  BlobX_BCCipher      cipher;      /* proc to en-/decrypt              */
  BlobX_BCClean       clean;       /* proc to remove dangerous info    */

  ClientData          clientData;  /* arbitrary context */
  Tcl_CmdDeleteProc*  cleanup;     /* proc to cleanup context */
  Opt_Table           optExt;      /* additional options */
  int                 optDataSize; /* additional required storage
				    * by 'optExt' handlers */
  Tcl_CmdDeleteProc*  cleanOpt;    /* proc to cleanup data collected
				      via 'optExt' */
} BcData;


/*
 * declaration of internal procedures
 */

static int
CipherCmd _ANSI_ARGS_ ((ClientData           clientData,
			BlobX_ScsOptionData* optInfo,
			Tcl_Interp*          interp,
			int                  argc,
			char**               argv));
static void
InitBC _ANSI_ARGS_ ((BlobX_ScsOptionData* opt,
		     ClientData           clientData));

static void
DeleteBC _ANSI_ARGS_ ((ClientData clientData));

static void
CleanOpt _ANSI_ARGS_ ((ClientData clientData));

/*
 * Option handler
 */

static int
HandleECB _ANSI_ARGS_ ((ClientData  clientData,
			Tcl_Interp* interp,
			CONST char* name,
			int         argc,
			char**      argv,
			int*        processed));
static int
HandleCBC _ANSI_ARGS_ ((ClientData  clientData,
			Tcl_Interp* interp,
			CONST char* name,
			int         argc,
			char**      argv,
			int*        processed));
static int
HandleCFB _ANSI_ARGS_ ((ClientData  clientData,
			Tcl_Interp* interp,
			CONST char* name,
			int         argc,
			char**      argv,
			int*        processed));
static int
HandleOFB _ANSI_ARGS_ ((ClientData  clientData,
			Tcl_Interp* interp,
			CONST char* name,
			int         argc,
			char**      argv,
			int*        processed));
static int
HandleSchedule _ANSI_ARGS_ ((ClientData  clientData,
			     Tcl_Interp* interp,
			     CONST char* name,
			     int         argc,
			     char**      argv,
			     int*        processed));
static int
HandleIV _ANSI_ARGS_ ((ClientData  clientData,
		       Tcl_Interp* interp,
		       CONST char* name,
		       int         argc,
		       char**      argv,
		       int*        processed));
static int
HandleShift _ANSI_ARGS_ ((ClientData  clientData,
			  Tcl_Interp* interp,
			  CONST char* name,
			  int         argc,
			  char**      argv,
			  int*        processed));
static int
HandleKey _ANSI_ARGS_ ((ClientData  clientData,
			Tcl_Interp* interp,
			CONST char* name,
			int         argc,
			char**      argv,
			int*        processed));
static int
HandleKeySchedule _ANSI_ARGS_ ((ClientData  clientData,
				Tcl_Interp* interp,
				CONST char* name,
				int         argc,
				char**      argv,
				int*        processed));

/*
 * Generic mode handling
 */

static void
ExecECB _ANSI_ARGS_ ((BlobX_BcOptInfo* opt));

static void
ExecCBC _ANSI_ARGS_ ((BlobX_BcOptInfo* opt));

static void
ExecCFB _ANSI_ARGS_ ((BlobX_BcOptInfo* opt));

static void
ExecOFB _ANSI_ARGS_ ((BlobX_BcOptInfo* opt));


/* The next 2 procedures were taken from PGPTools:idea.c
 * Original names are 'cfbshift' and 'xorbuf':
 *
 * >>	Can be applied for any block encryption algorithm,
 * >>	with any block size, such as the DES or the IDEA cipher.
 */

static void
ShiftBuf _ANSI_ARGS_ ((char* buf, char* in,
		       int shiftWidth, int blockSize));

static void
XorBuf _ANSI_ARGS_ ((char* buf, char* in, int blockSize));


static void
TooShort _ANSI_ARGS_ ((Tcl_Interp* interp,
		       CONST char* name,
		       int         minSize));

/*
 *------------------------------------------------------*
 *
 *	BlobX_BCOptions --
 *
 *	------------------------------------------------*
 *	Retrieves the table of options known by block ciphers.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Memory is allocated.
 *
 *	Result:
 *		The created Opt_Table
 *
 *------------------------------------------------------*
 */

Opt_Table
BlobX_BCOptions (interp)
Tcl_Interp* interp;
{
  static Opt_Table bcOptions = 0;

  if (! bcOptions)
    {
      static Opt_Spec bc [] =
	{
	  {"keyschedule",  HandleKeySchedule, "define internal key to use"},
	  {"key",          HandleKey,         "define key to use"},
	  /* 'key' overloads definition in SCS */

	  {"schedule-key", HandleSchedule, "request precompilation of key"},

	  {"ecb",          HandleECB,      "request electronic code book"},
	  {"cbc",          HandleCBC,      "request cipher block chaining"},
	  {"cfb",          HandleCFB,      "request cipher feedback"},
	  {"ofb",          HandleOFB,      "request output feedback"},

	  {"iv",           HandleIV,     "define initialization vector"},
	  {"shift",        HandleShift,  "define shiftwidth of feedback modes"}
	};

      bcOptions = Opt_CompileArray (interp,
				     sizeof (bc) / sizeof (Opt_Spec),
				     bc);
      assert (bcOptions);

      /* include standard options of SCS */
      Opt_LinkTables (bcOptions, BlobX_SCSOptions (interp));
    }

  return bcOptions;
}

/*
 *------------------------------------------------------*
 *
 *	BlobX_RegisterBC --
 *
 *	------------------------------------------------*
 *	The procedure creates a new command with name
 *	'name'.  At every successful invocation the
 *	internal general block-cipher procedure is
 *	executed to en-/decrypt the provided data with
 *	the specified key using the actual block-cipher.
 *
 *	See manual for calling syntax.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above. 
 *
 *	Result:
 *		a standard TCL error code.
 *
 *------------------------------------------------------*
 */

EXTERN int
BlobX_RegisterBC (interp, name, keySize, blockSize,
		  schedule, cipher, clean, clientData,
		  cleanup, optExt, optDataSize, cleanOpt)

Tcl_Interp*         interp;      /* interpreter to extend */
CONST char*         name;        /* name of new command */
int                 keySize;     /* expected size of key (in bytes) */
int                 blockSize;   /* size of single block (in bytes) */
BlobX_BCScheduleKey schedule;    /* proc to do key scheduling */
BlobX_BCCipher      cipher;      /* proc to doen-/decryption */
BlobX_BCClean       clean;       /* proc to remove dangerous info */
ClientData          clientData;  /* arbitrary context */
Tcl_CmdDeleteProc*  cleanup;     /* proc to cleanup context */
Opt_Table           optExt;      /* additional options */
int                 optDataSize; /* additional required storage by
				    'optExt' handlers */
Tcl_CmdDeleteProc* cleanOpt;     /* procedure to delete option data
				    collected via 'optExt'. allowed
				    to be NULL */
{
  BcData*          bcDataPtr = (BcData*) ckalloc (sizeof (BcData));
  Opt_Table        options;
  int              res;


  if (! bcDataPtr)
    {
      Tcl_AppendResult (interp,
			"not enough memory to register '",
			name, "'", 0);
      return TCL_ERROR;
    }

  options = (optExt ? optExt : BlobX_BCOptions (interp));

  bcDataPtr->keySize     = keySize;
  bcDataPtr->blockSize   = blockSize;
  bcDataPtr->schedule    = schedule;
  bcDataPtr->cipher      = cipher;
  bcDataPtr->clean       = clean;
  bcDataPtr->clientData  = clientData;
  bcDataPtr->cleanup     = cleanup;
  bcDataPtr->optExt      = options;
  bcDataPtr->optDataSize = optDataSize;
  bcDataPtr->cleanOpt    = cleanOpt;


  res =
    BlobX_RegisterSCS (interp, name, CipherCmd, InitBC,
		       (ClientData) bcDataPtr, DeleteBC, options,
		       sizeof (BlobX_BcOptionData) + optDataSize,
		       CleanOpt);

  if (res != TCL_OK)
    {
      ckfree (bcDataPtr);
      return res;
    }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	CipherCmd --
 *
 *	------------------------------------------------*
 *	This procedure realizes the generic part of BC
 *	operation (argument parsing and mode operation).
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		a standard TCL error code
 *
 *------------------------------------------------------*
 */

static int
CipherCmd (clientData, extOptInfo, interp, argc, argv)
ClientData           clientData; /* arbitrary information required
				    by actual function (defined at
				    registration time) */
BlobX_ScsOptionData* extOptInfo; /* information defined via options */
Tcl_Interp*          interp;     /* interpreter we are working in */
int                  argc;       /* number of trailing arguments   (UNUSED) */
char**               argv;       /* arguments with additional info (UNUSED) */
{
  /*
   * The generic SCS code handled the option processing before.
   * Now we are able to do additional checks. Main task however
   * is the execution of the requested cipher.
   */

  BlobX_BcOptInfo* optInfo = (BlobX_BcOptInfo*) extOptInfo;
  BcData*    bcDef         = (BcData*) clientData;
  int        modulus;


  if (optInfo->scs.direction == SCHEDULE)
    {
      /* required is '-key'.
       */

      if (optInfo->bc.keytype != KEY_BLOCK)
	{
	  Tcl_AppendResult (interp, "can't schedule a schedule", 0);
	  return TCL_ERROR;
	}
      else if (! optInfo->scs.key.data)
	{
	  Tcl_AppendResult (interp, "no key specified for schedule", 0);
	  return TCL_ERROR;
	}

      /* remove possibly defined '-data' */
      if (optInfo->scs.data.data)
	{
	  assert (optInfo->scs.data.isAlloc);
	  ckfree (optInfo->scs.data.data);
	}

      /* do schedule and store result in the
       * place expected by the SCS code
       */

      optInfo->scs.data.data = (*bcDef->schedule) (optInfo,
						   &optInfo->scs.data.length);
      optInfo->scs.data.isAlloc = TRUE;

      return TCL_OK;
    }

  /* encrypt / decrypt from now on:
   * general requirements:
   *
   *	-data
   *	-ecb, -cbc, -cfb or -ofb
   *	-key or -key-schedule 
   *
   * stream modes:   -iv
   * feedback modes: -iv and -shift
   *
   * all modes place additional constraints on the size of 'data'.
   */

  if (optInfo->bc.mode == NO_MODE)
    {
      Tcl_AppendResult (interp, "mode not defined", 0);
      return TCL_ERROR;
    }
  else if (optInfo->scs.data.data == 0)
    {
      Tcl_AppendResult (interp, "data missing", 0);
      return TCL_ERROR;
    }


  modulus = bcDef->blockSize;

  if (optInfo->bc.mode != ECB)
    {
      /* stream modes (CBC, CFB, OFB) */

      if (! optInfo->bc.iv.data)
	{
	  Tcl_AppendResult (interp, "init-vector not defined", 0);
	  return TCL_ERROR;
	}

      if (optInfo->bc.mode != CBC)
	{
	  /* feedback modes (CFB, OFB) */

	  if (optInfo->bc.shiftwidth == 0)
	    {
	      Tcl_AppendResult (interp, "shift not defined", 0);
	      return TCL_ERROR;
	    }

	  modulus = optInfo->bc.shiftwidth;
	}
    }

  /* modulus:
   * ECB, CBC: blocksize
   * CFB, OFB: shiftwidth
   */

  if (0 != (optInfo->scs.data.length % modulus))
    {
      char buf [50];
      sprintf (buf, "%d", modulus);

      Tcl_AppendResult (interp,
			"size of specified data is not a multiple of ",
			buf, " bytes", 0);
      return TCL_ERROR;
    }

#if 0 /* catched in SCS code already */
  if (! optInfo->scs.key.data)
    {
      Tcl_AppendResult (interp, "neither key nor schedule specified", 0);
      return TCL_ERROR;
    }
#endif

  /* checks (BC) done now.
   * Of course the block cipher is allowed to do additional checks.
   * These have to done in 'schedule'.
   *
   * -- start cipher execution with compilation of key into schedule.
   */


  if (optInfo->bc.keytype == KEY_BLOCK)
    {
      /* compute schedule
       */

      int   length;
      char* data;
      int   direction;

      direction = optInfo->scs.direction;

      /* feedback modes operate cipher in encryption mode on both sides.
       * the real choice is saved and restored later.
       */
      if ((optInfo->bc.mode == CFB) || (optInfo->bc.mode == OFB))
	optInfo->scs.direction = ENCRYPT;

      data = (*bcDef->schedule) (optInfo, &length);

      optInfo->scs.direction = direction;


      if (! data)
	return TCL_ERROR;

      if (optInfo->scs.key.isAlloc)
	ckfree (optInfo->scs.key.data);

      optInfo->scs.key.data    = data;
      optInfo->scs.key.length  = length;
      optInfo->scs.key.isAlloc = TRUE;
    }

  /* -- dispatch according to mode */

  switch (optInfo->bc.mode)
    {
    case ECB:
      ExecECB (optInfo);
      break;

    case CBC:
      ExecCBC (optInfo);
      break;

    case CFB:
      ExecCFB (optInfo);
      break;

    case OFB:
      ExecOFB (optInfo);
      break;

    default:
      /* should not happen */
      assert (0);
    }


  /* -- closing work */

  if (optInfo->bc.keytype == KEY_BLOCK)
    {
      /* a temporary key schedule was made, zap it now
       */

      if (bcDef->clean)
	(*bcDef->clean) (optInfo->scs.key.data);
      else
	memset (optInfo->scs.key.data, '\0', optInfo->scs.key.length);

      /* release is done by SCS code */
    }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	InitBC --
 *
 *	------------------------------------------------*
 *	Initialize block cipher specific option structure
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above.
 *
 *	Result:
 *		None.
 *
 *------------------------------------------------------*
 */

static void
InitBC (opt, clientData)
BlobX_ScsOptionData* opt;
ClientData           clientData;
{
  BlobX_BcOptInfo* optInfo = (BlobX_BcOptInfo*) opt;  
  optInfo->bc.def          = clientData;
}

/*
 *------------------------------------------------------*
 *
 *	DeleteBC --
 *
 *	------------------------------------------------*
 *	Frees all memory associated to a command
 *	realizing a BC
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above
 *
 *	Result:
 *		None
 *
 *------------------------------------------------------*
 */

static void
DeleteBC (clientData)
ClientData  clientData;
{
  ckfree ((BcData*) clientData);
}

/*
 *------------------------------------------------------*
 *
 *	CleanOpt --
 *
 *	------------------------------------------------*
 *	Free memory associated to BC option information
 *	and everything defined by additional cipher options.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above
 *
 *	Result:
 *		None.
 *
 *------------------------------------------------------*
 */

static void
CleanOpt (clientData)
ClientData  clientData;
{
  BlobX_BcOptInfo* optInfo = (BlobX_BcOptInfo*) clientData;
  BcData*    bcDef         = (BcData*)    optInfo->bc.def;

  /* scs.key, scs.data were handled before by SCS code.
   * Now handle initialization vector.
   */

  if (optInfo->bc.iv.isAlloc)
    ckfree (optInfo->bc.iv.data);

  if (bcDef->cleanOpt)
    (*bcDef->cleanOpt) (optInfo);
}

/*
 * Option handler
 */

/*
 *------------------------------------------------------*
 *
 *	HandleECB --
 *
 *	------------------------------------------------*
 *	process option '-ecb'. no additional arguments
 *	are required.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above
 *
 *	Result:
 *		a standard TCL error code
 *
 *------------------------------------------------------*
 */

static int
HandleECB (clientData, interp, name, argc, argv, processed)
ClientData  clientData;  /* arbitrary context */
Tcl_Interp* interp;      /* interpreter for messages */
CONST char* name;        /* name of option */
int         argc;        /* #arguments behind option */
char**      argv;        /* trailing arguments, possibly
			    arguments to option itself */
int*        processed;   /* number of processed option-
			    arguments, preset by caller to 0 */
{
  BlobX_BcOptInfo* optInfo = (BlobX_BcOptInfo*) clientData;

  optInfo->bc.mode = ECB;

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	HandleCBC --
 *
 *	------------------------------------------------*
 *	process option '-cbc'. no additional arguments
 *	are required.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above
 *
 *	Result:
 *		a standard TCL error code
 *
 *------------------------------------------------------*
 */

static int
HandleCBC (clientData, interp, name, argc, argv, processed)
ClientData  clientData;  /* arbitrary context */
Tcl_Interp* interp;      /* interpreter for messages */
CONST char* name;        /* name of option */
int         argc;        /* #arguments behind option */
char**      argv;        /* trailing arguments, possibly
			    arguments to option itself */
int*        processed;   /* number of processed option-
			    arguments, preset by caller to 0 */
{
  BlobX_BcOptInfo* optInfo = (BlobX_BcOptInfo*) clientData;

  optInfo->bc.mode = CBC;

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	HandleCFB --
 *
 *	------------------------------------------------*
 *	process option '-cfb'. no additional arguments
 *	are required.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above
 *
 *	Result:
 *		a standard TCL error code
 *
 *------------------------------------------------------*
 */

static int
HandleCFB (clientData, interp, name, argc, argv, processed)
ClientData  clientData;  /* arbitrary context */
Tcl_Interp* interp;      /* interpreter for messages */
CONST char* name;        /* name of option */
int         argc;        /* #arguments behind option */
char**      argv;        /* trailing arguments, possibly
			    arguments to option itself */
int*        processed;   /* number of processed option-
			    arguments, preset by caller to 0 */
{
  BlobX_BcOptInfo* optInfo = (BlobX_BcOptInfo*) clientData;

  optInfo->bc.mode = CFB;

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	HandleOFB --
 *
 *	------------------------------------------------*
 *	process option '-ofb'. no additional arguments
 *	are required.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above
 *
 *	Result:
 *		a standard TCL error code
 *
 *------------------------------------------------------*
 */

static int
HandleOFB (clientData, interp, name, argc, argv, processed)
ClientData  clientData;  /* arbitrary context */
Tcl_Interp* interp;      /* interpreter for messages */
CONST char* name;        /* name of option */
int         argc;        /* #arguments behind option */
char**      argv;        /* trailing arguments, possibly
			    arguments to option itself */
int*        processed;   /* number of processed option-
			    arguments, preset by caller to 0 */
{
  BlobX_BcOptInfo* optInfo = (BlobX_BcOptInfo*) clientData;

  optInfo->bc.mode = OFB;

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	HandleSchedule --
 *
 *	------------------------------------------------*
 *	process option '-schedule-key'. no additional
 *	arguments are required.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above
 *
 *	Result:
 *		a standard TCL error code
 *
 *------------------------------------------------------*
 */

static int
HandleSchedule (clientData, interp, name, argc, argv, processed)
ClientData  clientData;  /* arbitrary context */
Tcl_Interp* interp;      /* interpreter for messages */
CONST char* name;        /* name of option */
int         argc;        /* #arguments behind option */
char**      argv;        /* trailing arguments, possibly
			    arguments to option itself */
int*        processed;   /* number of processed option-
			    arguments, preset by caller to 0 */
{
  BlobX_BcOptInfo* optInfo = (BlobX_BcOptInfo*) clientData;

  optInfo->scs.direction = SCHEDULE;

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	HandleIV --
 *
 *	------------------------------------------------*
 *	process option '-iv'. the following 2-5 arguments
 *	have to follow the syntax accepted by 'Blob_GetInput'.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above
 *
 *	Result:
 *		a standard TCL error code
 *
 *------------------------------------------------------*
 */

static int
HandleIV (clientData, interp, name, argc, argv, processed)
ClientData  clientData;  /* arbitrary context */
Tcl_Interp* interp;      /* interpreter for messages */
CONST char* name;        /* name of option */
int         argc;        /* #arguments behind option */
char**      argv;        /* trailing arguments, possibly
			    arguments to option itself */
int*        processed;   /* number of processed option-
			    arguments, preset by caller to 0 */
{
  BlobX_BcOptInfo* optInfo = (BlobX_BcOptInfo*) clientData;
  BcData*    bcDef         = (BcData*)    optInfo->bc.def;
  int        res;

  int   length;
  int   isAlloc;
  char* data;


  if (argc < 2)
    {
      Tcl_AppendResult (interp, "-iv without enough arguments", 0);
      return TCL_ERROR;
    }


  res = Blob_GetInput (interp, argc, argv, 0, "-iv",
		       &length, &data, &isAlloc, processed);
  if (res != TCL_OK)
    {
      Tcl_AppendResult (interp, " (-iv)", 0);
      return res;
    }


  /* preliminary checks */

  if (length < bcDef->blockSize)
    {
      TooShort (interp, "iv", bcDef->blockSize);
      return TCL_ERROR;
    }


  /* store data for use later on */

  if (optInfo->bc.iv.data)
    {
      /* there was an earlier definition, reuse occupied space */
      assert (optInfo->bc.iv.isAlloc);

      memcpy ((VOID*) optInfo->bc.iv.data,
	      (VOID*) data,
	      bcDef->blockSize);

      if (isAlloc)
	ckfree (data);
    }
  else if (! isAlloc)
    {
      /* Make a copy of the IV. This is required to
       * avoid overwriting the source (which might be a blob).
       * Superfluous bytes are cut here too.
       */

      char* tmp = ckalloc (bcDef->blockSize);
      
      if (! tmp)
	{
	  Tcl_AppendResult (interp, "(-iv) not enough memory", 0);
	  return TCL_ERROR;
	}

      memcpy ((VOID*) tmp, (VOID*) data, bcDef->blockSize);

      data    = tmp;
      length  = bcDef->blockSize;
      isAlloc = TRUE;
    }
  /* else isAlloc: we can use info directly */

  optInfo->bc.iv.isAlloc = isAlloc;
  optInfo->bc.iv.data    = data;
  optInfo->bc.iv.length  = length;

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	HandleShift --
 *
 *	------------------------------------------------*
 *	process option '-shift'. a single integer number
 *	is required.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above
 *
 *	Result:
 *		a standard TCL error code
 *
 *------------------------------------------------------*
 */

static int
HandleShift (clientData, interp, name, argc, argv, processed)
ClientData  clientData;  /* arbitrary context */
Tcl_Interp* interp;      /* interpreter for messages */
CONST char* name;        /* name of option */
int         argc;        /* #arguments behind option */
char**      argv;        /* trailing arguments, possibly
			    arguments to option itself */
int*        processed;   /* number of processed option-
			    arguments, preset by caller to 0 */
{
  BlobX_BcOptInfo* optInfo = (BlobX_BcOptInfo*) clientData;
  BcData*    bcDef         = (BcData*)    optInfo->bc.def;
  int sw, res;

  if (argc < 1)
    {
      Tcl_AppendResult (interp, "-shift without argument", 0);
    }

  /* a temporary ('sw') is used to preserve an already defined
   * value as long as possible
   */

  res = Tcl_GetInt (interp, argv [0], &sw);
  if (res != TCL_OK)
    {
      Tcl_AppendResult (interp, " (-shift)", 0);
      return res;
    }


  /* preliminary checks */

  if (sw > bcDef->blockSize)
    {
      Tcl_AppendResult (interp,
			"shiftvalue greater than size of blocks", 0);
      return TCL_ERROR;
    }
  else if (sw <= 0)
    {
      Tcl_AppendResult (interp,
			"illegal value <= 0 presented to -shift", 0);
      return TCL_ERROR;
    }
  else if (0 != (bcDef->blockSize % sw))
    {
      Tcl_AppendResult (interp,
			"shiftvalue is no divisor of blocksize", 0);
      return TCL_ERROR;
    }

  /* accept new value */

  optInfo->bc.shiftwidth = sw;
  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	HandleKey --
 *
 *	------------------------------------------------*
 *	process option '-key'. the following 2-5 arguments
 *	have to follow the syntax accepted by 'Blob_GetInput'.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above.
 *
 *	Result:
 *		a standard TCL error code
 *
 *------------------------------------------------------*
 */

static int
HandleKey (clientData, interp, name, argc, argv, processed)
ClientData  clientData;  /* arbitrary context */
Tcl_Interp* interp;      /* interpreter for messages */
CONST char* name;        /* name of option */
int         argc;        /* #arguments behind option */
char**      argv;        /* trailing arguments, possibly
			    arguments to option itself */
int*        processed;   /* number of processed option-
			    arguments, preset by caller to 0 */
{
  BlobX_BcOptInfo* optInfo = (BlobX_BcOptInfo*) clientData;
  BcData*    bcDef         = (BcData*)    optInfo->bc.def;
  int        res;

  int   length;
  int   isAlloc;
  char* data;

  if (argc < 2)
    {
      Tcl_AppendResult (interp, "-key without enough arguments", 0);
      return TCL_ERROR;
    }

  res = Blob_GetInput (interp, argc, argv,
		       0, "-key",
		       &length, &data, &isAlloc, processed);

  if (res != TCL_OK)
    {
      Tcl_AppendResult (interp, " (-key)", 0);
      return res;
    }

  /* preliminary checks */

  if (length < bcDef->keySize)
    {
      TooShort (interp, "key", bcDef->keySize);
      return TCL_ERROR;
    }


  if (optInfo->scs.key.data && optInfo->scs.key.isAlloc)
    ckfree (optInfo->scs.key.data);

  optInfo->scs.key.length  = length;
  optInfo->scs.key.data    = data;
  optInfo->scs.key.isAlloc = isAlloc;

  optInfo->bc.keytype      = KEY_BLOCK;

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	HandleKeyschedule --
 *
 *	------------------------------------------------*
 *	process option '-key-schedule'. the following 2-5
 *	arguments have to follow the syntax accepted by
 *	'Blob_GetInput'.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above.
 *
 *	Result:
 *		a standard TCL error code
 *
 *------------------------------------------------------*
 */

static int
HandleKeySchedule (clientData, interp, name, argc, argv, processed)
ClientData  clientData;  /* arbitrary context */
Tcl_Interp* interp;      /* interpreter for messages */
CONST char* name;        /* name of option */
int         argc;        /* #arguments behind option */
char**      argv;        /* trailing arguments, possibly
			    arguments to option itself */
int*        processed;   /* number of processed option-
			    arguments, preset by caller to 0 */
{
  BlobX_BcOptInfo* optInfo = (BlobX_BcOptInfo*) clientData;
  int        res;

  int   length;
  int   isAlloc;
  char* data;

  if (argc < 2)
    {
      Tcl_AppendResult (interp, "-key-schedule without enough arguments", 0);
      return TCL_ERROR;
    }

  res = Blob_GetInput (interp, argc, argv,
		       0, "-key",
		       &length, &data, &isAlloc, processed);

  if (res != TCL_OK)
    {
      Tcl_AppendResult (interp, " (-key-schedule)", 0);
      return res;
    }

  if (optInfo->scs.key.data && optInfo->scs.key.isAlloc)
    ckfree (optInfo->scs.key.data);

  optInfo->scs.key.length  = length;
  optInfo->scs.key.data    = data;
  optInfo->scs.key.isAlloc = isAlloc;

  optInfo->bc.keytype      = KEY_SCHED;

  return TCL_OK;
}

/*
 * Generic code of cipher modes.
 */

/*
 *------------------------------------------------------*
 *
 *	ExecECB --
 *
 *	------------------------------------------------*
 *	Implementation of generic mode ECB ==  Electronic CodeBook.
 *	All blocks are handled independently.  This is
 *	an insecure mode.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		None.
 *
 *------------------------------------------------------*
 */

static void
ExecECB (opt)
BlobX_BcOptInfo* opt;
{
  BcData*    bcDef  = (BcData*) opt->bc.def;
  int        blocks = opt->scs.data.length / bcDef->blockSize;
  char*      data   = opt->scs.data.data;
  int        i;

  for (i=0; i < blocks; i++, data += bcDef->blockSize)
    {
      (*bcDef->cipher) (opt->scs.key.data, data, data);
    }
}

/*
 *------------------------------------------------------*
 *
 *	ExecCBC --
 *
 *	------------------------------------------------*
 *	Implementation of generic mode CBC == Cipher Block Chaining.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		None.
 *
 *------------------------------------------------------*
 */

static void
ExecCBC (opt)
BlobX_BcOptInfo* opt;
{
  BcData*    bcDef  = (BcData*) opt->bc.def;
  int        blocks = opt->scs.data.length / bcDef->blockSize;
  char*      data   = opt->scs.data.data;
  int        i;

  if (opt->scs.direction == ENCRYPT)
    {
      for (i=0; i < blocks; i++, data += bcDef->blockSize)
	{
	  XorBuf (data, opt->bc.iv.data, bcDef->blockSize);

	  (*bcDef->cipher) (opt->scs.key.data, data, data);
	  memcpy ((VOID*) opt->bc.iv.data, (VOID*) data, bcDef->blockSize);
	}
    }
  else
    {
      for (i=0; i < blocks; i++, data += bcDef->blockSize)
	{
	  (*bcDef->cipher) (opt->scs.key.data, data, data);

	  XorBuf (data, opt->bc.iv.data, bcDef->blockSize);
	  memcpy ((VOID*) opt->bc.iv.data, (VOID*) data, bcDef->blockSize);
	}
    }
}

/*
 *------------------------------------------------------*
 *
 *	ExecCFB --
 *
 *	------------------------------------------------*
 *	Implementation of generic mode CFB == Cipher FeedBack.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		None.
 *
 *------------------------------------------------------*
 */

static void
ExecCFB (opt)
BlobX_BcOptInfo* opt;
{
  BcData*    bcDef  = (BcData*) opt->bc.def;
  int        blocks = opt->scs.data.length / opt->bc.shiftwidth;
  char*      data   = opt->scs.data.data;
  int        i;

  /* create temporary register for feedback */
  char* tmp = ckalloc (bcDef->blockSize);
  assert (tmp);

  if (opt->scs.direction == ENCRYPT)
    {
      for (i=0; i < blocks; i++, data += opt->bc.shiftwidth)
	{
	  (*bcDef->cipher) (opt->scs.key.data, opt->bc.iv.data, tmp);

	  XorBuf   (data,            tmp,  opt->bc.shiftwidth);
	  ShiftBuf (opt->bc.iv.data, data, opt->bc.shiftwidth, bcDef->blockSize);
	}
    }
  else
    {
      for (i=0; i < blocks; i++, data += opt->bc.shiftwidth)
	{
	  (*bcDef->cipher) (opt->scs.key.data, opt->bc.iv.data, tmp);

	  ShiftBuf (opt->bc.iv.data, data, opt->bc.shiftwidth, bcDef->blockSize);
	  XorBuf   (data,            tmp,  opt->bc.shiftwidth);
	}
    }

  ckfree (tmp);
}

/*
 *------------------------------------------------------*
 *
 *	ExecOFB --
 *
 *	------------------------------------------------*
 *	Implementation of generic mode OFB == Output FeedBack.
 *	It is said to be insecure for a shiftwidth
 *	smaller than the blocksize.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		None.
 *
 *------------------------------------------------------*
 */

static void
ExecOFB (opt)
BlobX_BcOptInfo* opt;
{
  BcData*    bcDef  = (BcData*) opt->bc.def;
  int        blocks = opt->scs.data.length / bcDef->blockSize;
  char*      data   = opt->scs.data.data;
  int        i;

  /* create temporary register for feedback */
  char* tmp = ckalloc (bcDef->blockSize);
  assert (tmp);

  for (i=0; i < blocks; i++, data += opt->bc.shiftwidth)
    {
      (*bcDef->cipher) (opt->scs.key.data, opt->bc.iv.data, tmp);

      XorBuf   (data,            tmp, opt->bc.shiftwidth);
      ShiftBuf (opt->bc.iv.data, tmp, opt->bc.shiftwidth, bcDef->blockSize);
    }

  ckfree (tmp);
}

/*
 *------------------------------------------------------*
 *
 *	ShiftBuf --
 *
 *	------------------------------------------------*
 *	Take the 'shiftWidth' leftmost bytes of 'in' and
 *	shift them into the rightmost bytes of 'buf'.
 *	The leftmost bytes of 'buf' are lost.  Both
 *	buffers are assumed to be of size 'blockSize'.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above.
 *
 *	Result:
 *		None.
 *
 *------------------------------------------------------*
 */

static void
ShiftBuf (buf, in, shiftWidth, blockSize)
char* buf;        /* register to shift */
char* in;         /* data to shift into the register */
int   shiftWidth; /* number of bytes to shift in */
int   blockSize;  /* size of both registers */
{
  assert (shiftWidth > 0);

  if (shiftWidth == blockSize)
    {
      /* special case of dropping the whole old register contents
       * added by me (aku).
       */
      memcpy (buf, in, blockSize);
    }
  else
    {
      int retained;

      /* number bytes in 'buf' to retain */
      retained = blockSize - shiftWidth;

      /* left-shift retained bytes of 'buf' over by
       * 'shiftWidth' bytes to make room
       */
      while (retained --)
	{
	  *buf = *(buf + shiftWidth);
	  buf ++;
	}

      /* now copy 'shiftwidth' bytes from 'input' to shifted tail of 'buf' */
      do
	{
	  *buf++ = *in++;
	}
      while (--shiftWidth);
    }
}

/*
 *------------------------------------------------------*
 *
 *	XorBuf --
 *
 *	------------------------------------------------*
 *	Do an exclusive or of all 'blockSize' bytes of
 *	'buf' with the corresponding bytes in 'mask'.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See above.
 *
 *	Result:
 *		None
 *
 *------------------------------------------------------*
 */

static void
XorBuf (buf, mask, blockSize)
char* buf;       /* register to manipulate */
char* mask;      /* data to mix with the register */
int   blockSize; /* size of both registers */
{
  assert (blockSize > 0);

  do
    {
      *buf++ ^= *mask++;
    }
  while (--blockSize);
}

/*
 *------------------------------------------------------*
 *
 *	TooShort --
 *
 *	------------------------------------------------*
 *	Generate errormessage for data being shorter than
 *	required.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		The interpreter result is changed.
 *
 *	Result:
 *		None.
 *
 *------------------------------------------------------*
 */

static void
TooShort (interp, name, minSize)
Tcl_Interp* interp;
CONST char* name;
int         minSize;
{
  char buf [50];
  sprintf (buf, "%d", minSize);

  Tcl_AppendResult (interp,
		    "length of specified ", name, " shorter than ",
		    buf, " bytes", 0);
}

